package PureMessage::Reports::Query::Remote;
use strict;
use warnings;
use Carp;
use PureMessage::ServerGroup qw(read_config get_hosts get_host get_group);
use PureMessage::RPC::Client;
use PureMessage::Reports::Query::GroupBy;
use HTTP::Date;
use Data::Dumper;

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    return bless {}, $class;
}

sub table_fields  { shift->{'table_fields'} }
sub result_fields { shift->{'result_fields'} }
sub results       { shift->{'results'} }

sub prepare {
    my $self = shift;
    my %ops = @_;
    
    $self->{'table_name'} = $ops{'table_name'} ||
	croak 'table_name is a required parameter';
	
    $self->{'remote_ds'} = $ops{'remote_ds'} || 
	croak "remote_ds is a required parameter";

    $self->{'time_frame'} = $ops{'time_frame'} ||
	croak 'time_frame is a required parameter';
    
    my @fields = ();
    if ($ops{'group'}) {
	$self->{'qspec'} = $ops{'group'};
	foreach my $spec (@{$self->{'qspec'}}) {
	    push(@fields, $$spec{'field_name'});
	}
    }
    elsif ($ops{'fields'}) {
	$self->{'fspec'} = $ops{'fields'};
	@fields = @{$ops{'fields'}};
    }
    else {
	croak "remote query prepare requires either 'group' or 'fields' to be specified";
    }
    $self->{'table_fields'} = \@fields;
    $self->{'result_fields'} = \@fields;

    my $start_at = str2time($ops{'start_at'}) if $ops{'start_at'};
    my $end_at   = str2time($ops{'end_at'})   if $ops{'end_at'};
    if ($end_at && $start_at) {
	croak "End time must be greater than or equal to the start time."
	    unless ($end_at >= $start_at);
    }
    $self->{'start_at'} = $ops{'start_at'} if $ops{'start_at'};
    $self->{'end_at'}   = $ops{'end_at'}   if $ops{'end_at'};

    my $cfg = read_config();    
    my ($type, $name) = split(':', $self->{'remote_ds'}, 2);
    my $hosts = 0;
    if ($type eq 'group') {
	croak "No such group '$name'" unless get_group($name);
	$hosts = get_hosts($name);
    }
    else {
	croak "No such host '$name'" unless get_host($name);
	$hosts = get_hosts([$name]);
    }	
    croak "No such group or host" unless $hosts;
    $self->{'hosts'} = $hosts;
    my @qry_params = $self->_prepare_marshaled_query();
    $self->{'qry_params'} = \@qry_params;
}

sub execute {
    my $self = shift;
    my $raw = $self->_marshal_queries();
    my @sorted = sort { $a->{'time_stamp'} cmp $b->{'time_stamp'} } @$raw;

    # if there are multiple hosts which created the resuls then the grouping
    # needs to be performed locally for the data to be meaningful.
    unless (scalar keys %{$self->{'hosts'}} == 1) {
	my $grp = PureMessage::Reports::Query::GroupBy->new('query_spec' => $self->{'qspec'});
	foreach my $row (@sorted) {
	    $grp->add_row($self->_mk_array_row($row));
	}
	$self->{'results'} = $grp;
    }
    else {
	my $rec = PureMessage::Reports::RPCRecordSet->new(
	    'data' => \@sorted,
	    'fields' => $self->{'table_fields'},
	    );
	$self->{'results'} = $rec;
    }
}

sub _mk_array_row {
    my $self = shift;
    my $hrow = shift;
    my @row = ();
    foreach my $f (@{$self->table_fields}) {
	push(@row, $hrow->{$f} || 0);	
    }
    return \@row;
}

# RPC does not yet support nested structures so we need to flatten the query 
#spec into an array in the structure that the receiver understands (list-based
# key/value). See PM::Manager::RPCResponse::Reports for the structure 
# definition.
sub _prepare_marshaled_query {
    my $self = shift;
    my @qry = (
	$self->{'time_frame'},
	'table_name', 
	$self->{'table_name'}
    );
    my $qspec = $self->{'qspec'};

    push(@qry, ('start_at', $self->{'start_at'})) if $self->{'start_at'};
    push(@qry, ('end_at', $self->{'end_at'})) if $self->{'end_at'};

    # Only send the groupby information if the query is going out to
    # a single host. Otherwise the grouping needs to be done locally to
    # include the results from all hosts properly.
    if (scalar keys %{$self->{'hosts'}} == 1) {
	my $cnt = 0;
	foreach my $spec (@$qspec) {
	    $cnt++;
	    my @item = (
		'groupby',    $cnt,
		'field_name', $spec->{'field_name'},
		'function',   $spec->{'function'},
	    );
	    push(@item, ('date_part', $spec->{'date_part'})) if $spec->{'date_part'};
	    push(@qry, @item);
	}
    }
    return @qry if wantarray;
    return \@qry;
}

sub _marshal_queries {
    my $self = shift;
    my $rpc = PureMessage::RPC::Client->new($self->{'hosts'});
    $rpc->call('Reports.remote_query', @{$self->{'qry_params'}});
    $rpc->wait();
    my @raw = ();
    foreach my $h (keys %{$self->{'hosts'}}) {
	if (my $err = $rpc->error($h)) {
	    $self->push_error($h, $err);
	}
	else {
	    my $data = $rpc->data($h);
            push(@raw,$data) if $data && ref($data) eq 'HASH';
    	    push(@raw, @$data) if $data && ref($data) eq 'ARRAY';
	}
    }
    return \@raw;
}

sub push_error {
    my ($self, $host, $err) = @_;
    push(@{$self->{'errors'}}, {$host, $err});
}

sub errors { 
    my $self = shift;
    return unless $self->{errors};
    return @{$self->{'errors'}} if wantarray;
    return $self->{'errors'};
}

sub has_errors { return scalar shift->errors }


package PureMessage::Reports::RPCRecordSet;
use strict;
use warnings;
use Carp;

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my %data = @_;
    my $self = bless \%data, $class;

    $self->{'rowp'} = 0;
    croak 'Missing data' unless $self->{'data'};
    croak 'Missing fields' unless $self->{'fields'};
    return $self;
}

sub fetchrow_arrayref {
    my $self = shift;
    return undef if $self->{'rowp'} >= scalar @{$self->{'data'}};
    my $row = $self->{'data'}->[$self->{'rowp'}++];
    $row = $self->_mk_array($row) if ref($row) eq 'HASH';
    return $row;    
}

sub fetchrow_hashref {
    my $self = shift;
    return undef if $self->{'rowp'} >= scalar @{$self->{'data'}};
    my $row = $self->{'data'}->[$self->{'rowp'}++];
    return $row;    
}

sub _mk_array {
    my ($self, $h) = @_;
    my @row;
    foreach my $f (@{$self->{'fields'}}) {
	push(@row, $h->{$f});
    }
    return \@row;
}

1;

