package PureMessage::ServerGroup::Publication;
use PureMessage::ServerGroup qw (:all);
use ActiveState::Config qw(:path);
use PureMessage::Version qw($PREFIX);
use PureMessage::RPC::Client;
use Digest::MD5;
use PureMessage::_sys ();
use PureMessage::List;
use PureMessage::Map;
use strict;

my $p;     # publication parser
my $l;     # list parser
my $m;     # map parser
my %conf;  # cached data
my %lconf; # cached list data
my %mconf; # cached map data
my $file = "$PREFIX/etc/publications.conf"; # default path, can be changed in new()
my $error;
my $hostname = PureMessage::_sys::gethostname();
my $writable = 0; # I don't understand this variable, seems counter-intuitive
my @listmeta = qw(name description precious source match_type comparator);
my @mapmeta  = qw(name description source match_type);

sub get_all {
    my %opts = @_;
    _read_config(file=>$opts{file}) unless %conf;
    my @all;
    foreach ( keys %{$conf{publication}} ) {
        push @all, PureMessage::ServerGroup::Publication->new($_);
    }
    return @all;
}

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $name = shift;
    my %opts = @_;
    my $self;
    _read_config(file=>$opts{file}) unless %conf;
    _read_meta_config(list=>$opts{list},map=>$opts{map});

    if ( ! $conf{publication}{$name} ) {
        if ($opts{create}) {
            # add new publication
            _add_publication($name, \%opts);
        }
        else {
            raise_error(undef, "Publication '$name' not found");
            return undef;
        }
    }
    $self->{name} = $name;
    $self->{error} = "";
    _init($self);
    return bless($self, $class);
}

sub _init {
    my $self = shift;
    my $p = $conf{publication}{$self->{name}};
    while ( my ($k, $v) = each %{$p} ) {
        $self->{$k} = $v;
	# load $self->{share} with meta-info for lists and maps
	if ( ($k eq "share") and (ref($v) eq "HASH") ) {
	    _init_obj($self, $v);
	}
    }
}

# load any %lconf and %mconf info into %conf ( list and map information )
sub _init_obj {
    my $self = shift;
    my $ref = shift;
    if ( %lconf ) {
	foreach ( keys %{$ref->{list}} ) {
	    $ref->{list}{$_} = $lconf{list}{$_} if exists($lconf{list}{$_});
	}
    }
    if ( %mconf ) {
	foreach ( keys %{$ref->{map}} ) {
	    $ref->{map}{$_} = $mconf{map}{$_} if exists($mconf{map}{$_});
	}
    }
}

sub _add_publication {
    my $name = shift;
    my $opts = shift;
    _read_config(writable=>1) unless $writable;
    my $path = "publication#$name";
    my $node = path_node($p->root, $path, create => 'group');
    unless (defined($node) && ref($node) =~ /^ActiveState::Config/) {
        raise_error(undef, "Can't create publication $name: $!");
        return;
    }
    foreach ('description') {
        my $opt = path_node($p->root, "$path.$_", create => 'value');
        $opt->value($opts->{$_});
        $node->attach($_, $opt);
    }
    # add created date
    path_node($p->root, "$path.created", create => 1);
    path_value($p->root, "$path.created", _today());
    # add empty <share> section
    path_node($p->root, "$path.share", create => 'group');
    # add empty <hosts> section
    path_node($p->root, "$path.hosts", create => 'group');
    save();
}

sub _today {
    my($sec,$min,$hour,$mday,$mon,$year) = localtime();
    return sprintf("%04d-%02d-%02dT%02d:%02d:%02d%s",
            $year+1900, $mon+1, $mday, $hour, $min, $sec, "");
}

sub _read_config {
    my %opts = @_;
    $file = $opts{file} if $opts{file};
    $writable = $opts{writable} || 0;
    $p = ActiveState::Config->new();
    $p->typedecl("publication#*.description", ['#str']);
    $p->typedecl("publication#*.share.file#*.*", ['#str']);
    $p->typedecl("publication#*.share.list#*.*", ['#str']);
    $p->typedecl("publication#*.share.map#*.*", ['#str']);
    $p->typedecl("publication#*.hosts", "#list");
    $p->parse_file($file, writable=>$writable, timeout=>5 );
    %conf = $p->as_hash();
}

# read lists.conf and maps.conf and parse info into %lconf and %mconf
sub _read_meta_config {
    my %opts = @_;
    # location of list and map configs hidden by PM::List and PM::Map
    $l = PureMessage::List->Config();
    $m = PureMessage::Map->Config();
    %lconf = $l->as_hash();
    %mconf = $m->as_hash();
    # set obj_type
    $lconf{list}{$_}{obj_type} = 'list' foreach ( keys %{$lconf{list}} );
    $mconf{map}{$_}{obj_type}  = 'map'  foreach ( keys %{$mconf{map}} );
    # set the path for all lists/maps with "source"
    for my $confdat ( $lconf{list}, $mconf{map} ) {
	foreach ( keys %{$confdat}  ) {
	    # did we do this already?
	    next if exists($confdat->{$_}{path}); 
	    my ($type, $path) = split ":", $confdat->{$_}{source};
	    if ( $type eq "file" ) {
		if ( -f "$PREFIX/etc/$path" && ! -f "$PREFIX/$path" ) {
		    $path = "etc/$path";
		}
		$confdat->{$_}{path} = $path;
	    }
	}
    }
}

sub rename {
    my $self = shift;
    my $name = shift;
    _read_config(writable=>1) unless $writable;
    my $path = "publication#$self->{name}";
    if ( my $node = path_node($p->root, $path) ) {
        $node->name($name);
        $self->{name} = $name;
        $self->save();
    }
}

sub delete {
    my $self = shift;
    my $path = "publication#$self->{name}";
    _read_config(writable=>1) unless $writable;
    path_nodes($p->root, $path, delete => 1);
    save();
}

sub attrib {
    my $self = shift;
    my $key = shift;
    my $val = shift;
    _read_config(writable=>1) unless $writable;
    unless ($val) {
        return $self->{$key};
    }
    my $path = "publication#$self->{name}";
    if ( my $node = path_node($p->root, $path) ) {
        path_node($p->root, "$path.$key", create => 'value');
        $val =~ s/[\r\n]/ /g;
        path_value($p->root, "$path.$key", $val);
        $self->{$key} = $val;
        $self->save();
        return 1;
    }
    return 0; 
}

sub subscribers {
    my $self = shift;
    my %opts = @_;
    $self->{hosts} ||= [];
    my $hosts = get_hosts($self->{hosts});
    if ( $opts{verify} ) {
        $self->verify($hosts);
    }
    return $hosts;
}

sub verify {
    my $self = shift;
    my $hosts = shift;
    my @f;
    # note that $hosts->{$host}{status}, $hosts->{$host}{files} used by publications.cgi
    # we want to query remote hosts and compare file, list and map signatures
    my $lists = $self->files(obj_type=>"list"); # $self->{share}{list}
    my $maps  = $self->files(obj_type=>"map");  # $self->{share}{map}
    my $files = $self->files(obj_type=>"file"); # $self->{share}{file}
    push @f, 'list:' . $lists->{$_}{path} foreach (keys %{$lists});
    push @f, 'map:'  . $maps->{$_}{path}  foreach (keys %{$maps});
    push @f, 'file:' . $files->{$_}{path} foreach (keys %{$files});

    unless ( @f ) {
        foreach my $host ( keys %{$hosts} ) {
            $hosts->{$host}{status} = "No files, lists or maps shared";
	    # $hosts->{$host}{status} is used in publications.cgi
        }   
        return;
    }
    my $rpc = PureMessage::RPC::Client->new( $hosts );

    $rpc->call("Subscription.Status", @f);
    $rpc->wait();
    foreach my $host ( keys %{$hosts} ) {
        $hosts->{$host}{status}   = "Ok"; # assume we are Ok unless told otherwise
        $hosts->{$host}{off_sync} = {};
        if ( my $err = $rpc->error($host) ) {
            $self->raise_error($err);
            $hosts->{$host}{status} = $err;
	    next if ( $err =~ /^Can\'t connect/ );
            next if ( $err =~ /^Authentication failed/ );
        }
	my %remote_objs = map { (ref($_) eq 'HASH') ? ($_->{path}=>$_) : (ERR=>$_) } $rpc->data($host);
	# This just gets the last non-hash ref
	if ( exists($remote_objs{ERR}) ) {
	    my $err = "Bad data reference returned for $host: " . delete($remote_objs{ERR});
	    $self->raise_error($err);
	    $hosts->{$host}{status} .= $err;
	}

	$hosts->{$host}{files} = \%remote_objs;
	# $hosts->{$host}{files} is used in publications.cgi

	foreach my $obj ( values %{$lists}, values %{$maps}, values %{$files} ) {
	    my $remote_obj = $remote_objs{$obj->{path}} || {}; # get handle on remote info
	    # check if we are in sync
	    if (  scalar(%$remote_obj) ) {
		$hosts->{$host}{off_sync}{$obj->{path}} = 1 
		    if ( ($remote_obj->{created} xor $obj->{created}) or
			 ($remote_obj->{signature} ne $obj->{signature}) );
	    }
	    my @metadata = ( $obj->{obj_type} eq "list" ) ? @listmeta : ();
	    @metadata = @mapmeta if ( $obj->{obj_type} eq "map" );
	    foreach ( @metadata ) {
		if ( exists($obj->{$_}) and (!defined($remote_obj->{$_}) or 
					     ($obj->{$_} ne $remote_obj->{$_})) ) {
		    if ( $_ eq "precious" ) {
			my ($rem, $loc) = ($remote_obj->{$_}, $obj->{$_});
			$rem = 0 if ( $rem && $rem =~ /^(no|false)$/i );
			$loc = 0 if ( $loc && $loc =~ /^(no|false)$/i );
			next unless ( $rem xor $loc );
		    }
		    $hosts->{$host}{off_sync}{$obj->{path}} = 1;
		}
	    }
	    # save date of most recent update on this host
	    $remote_obj->{modified} ||= 0;
	    # $hosts->{$host}{last_updated} is used in publications.cgi
	    $hosts->{$host}{last_updated} = $remote_obj->{modified}
	        unless ( defined($hosts->{$host}{last_updated}) and 
			 ($hosts->{$host}{last_updated} > $remote_obj->{modified}) );
	}

	$hosts->{$host}{last_updated} = 0; # reset !?
	my $os_files = scalar(keys %{$hosts->{$host}{off_sync}});
	if ( $os_files ) {
	    $hosts->{$host}{status} =~ s/^Ok$//; # not Ok anymore
	    $hosts->{$host}{status} .= " Out of sync: ";
	    $hosts->{$host}{status} .= "$os_files file(s)";
	}
    }
}

sub is_shared {
    my $file = shift;
    my $type = shift;
    my @type = defined($type) ? $type : qw(file list map);
    _read_config() unless %conf;
    _read_meta_config(); # unless ( %lconf && %mconf );
    foreach my $pub (keys %{$conf{publication}}) { # check each publication
	foreach my $t ( @type ) { # for these object types
	    foreach my $o ( keys %{$conf{publication}{$pub}{share}{$t}} ) { # foreach shared object 
		# peek at the path 
		if ( ($t eq "file") && exists($conf{publication}{$pub}{share}{file}{$o}) &&
		     ($conf{publication}{$pub}{share}{file}{$o}{path} eq $file) ) { 
		    return $pub;
		    # for lists and maps we peek
		} elsif ( ($t eq "list") && exists($lconf{list}{$o}) &&
			  ($lconf{list}{$o}{path} eq $file) ) {
		    return $pub;
		} elsif ( ($t eq "map") && exists($mconf{map}{$o}) &&
			  ($mconf{map}{$o}{path} eq $file) ) {
		    return $pub;
		}
	    }
	}
    }
    return 0;
}

sub add_file {
    my $self = shift;
    my $file_path = shift;
    my %opts = @_;
    my $obj_type = "file";
    if ( exists($opts{obj_type}) ) {
	$obj_type = $opts{obj_type};
    }
    _read_config(writable=>1) unless $writable;
    # load lists.conf into $l, maps.conf into $m
    _read_meta_config() if ( $obj_type ne "file" ); 
    if ( my $pub = is_shared($file_path, $obj_type) ) {
	$self->raise_error("$obj_type $file_path is part of the '$pub' publication.");
	return;
    }

    if ( ! -f "$PREFIX/$file_path" ) {
        $self->raise_error("Can't publish file $file_path: file not found.");
        return;
    }

    my $path = "publication#$self->{name}";
    my $uniq_name = $file_path; 
    $uniq_name =~ s/^etc\/*//;
    $uniq_name =~ s/[\s.\/]/_/g;

    my $node = path_node($p->root, "$path.share.$obj_type#$uniq_name", create => 'group');
    if ( $obj_type eq 'file' ) {
	$opts{path} = $file_path;
	foreach ('description', 'path') {
	    my $opt = path_node($p->root, "$path.share.file#$uniq_name.$_", create => 'value');
	    $opt->value($opts{$_});
	    $node->attach($_, $opt);
	}
    }
    save($self);
    $self->_init();
    return $uniq_name;
}

sub revoke_file {
    my $self = shift;
    my $uniq_name = shift;
    my %opts = @_;
    my $obj_type = "file";
    if ( exists($opts{obj_type}) ) {
	$obj_type = $opts{obj_type};
    }
    my $path = "publication#$self->{name}";
    _read_config(writable=>1) unless $writable;
    path_nodes($p->root, "$path.share.$obj_type#$uniq_name", delete => 1);
    save($self);
    delete $conf{publication}{$self->{name}}{share}{$obj_type}{$uniq_name};
    $self->_init($self->{name});
}

sub files {
    my $self = shift;
    my %opts = @_;
    my $obj_type = exists($opts{obj_type}) ? $opts{obj_type} : "file";
    if ( $obj_type eq "file" ) {
	$self->update_files_cache($obj_type);
    } elsif ( ($obj_type eq "list") || ($obj_type eq "map") ) {
	_read_config() unless %conf;
	_read_meta_config() unless ( %lconf && %mconf );
	$self->_init($self->{name}); # load meta-data into %conf
	$self->update_files_cache($obj_type);

    } else {
	die "Tried to get a list of bad data type: $obj_type"
    }
    return (exists($self->{share}{$obj_type})) ? $self->{share}{$obj_type} : {};
}

sub update_files_cache {
    my $self = shift;
    my $obj_type = shift;
    $obj_type ||= "file";
    my $files = $self->{share}{$obj_type};
    foreach ( keys %{$files} ) {
        next if $files->{$_}{signature}; # the information is already there, ignore request
	if ( $obj_type ne "file" ) {
	    my ($type, $source) = split ":", $files->{$_}{source};
	    next unless ( $type eq "file" ); # ignore unless source is a file
	    $source = "etc/$source" if ( -f "$PREFIX/etc/$source" and ! -f "$PREFIX/$source" );
	    $files->{$_}{path} = $source;
	}
        my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime) = stat($PREFIX."/".$files->{$_}{path});
        $files->{$_}{modified} = $mtime;
        $files->{$_}{created} = $ctime;
        $files->{$_}{size} = $size;
        $files->{$_}{signature} = $size ? _checksum($PREFIX."/".$files->{$_}{path}) : "None";
	$files->{$_}{obj_type} = $obj_type;
    }
}

sub _checksum {
    my $file = shift;
    my $ctx = Digest::MD5->new;
    open(my $F, $file);
    die "file is a directory: $file prefix = $PREFIX" if -d $file;
    $ctx->addfile($F);
    return $ctx->hexdigest;
}

sub sync {
    my $self = shift;
    my @h = @_; # all unless defined
    $self->{hosts} ||= [];
    @h = @{$self->{hosts}} unless scalar @h > 0;
    my $hosts = get_hosts(\@h);
    # XXX need to call verify() only if status information is missing
    $self->verify($hosts);
    while ( my ($host, $hostdata) = each %$hosts ) {
        next if ( scalar(keys %{$hosts->{$host}{off_sync}}) < 1 );
        my $rpc = PureMessage::RPC::Client->new( {$host=>$hostdata}, timeout=>60 );
	my %metainfo = ();
      FILE: foreach my $file ( keys %{$hosts->{$host}{off_sync}} )  {
	  # XXX
	  foreach my $list ( values %{$self->files(obj_type=>"list")} ) {
	      if ( $list->{path} eq $file ) {
		  $rpc->attach( "list" => [$PREFIX."/".$file, $file]  );
		  (my $listname = $list->{path}) =~ s/.*\/([\w\-]+)$/$1/;
		  foreach  ( @listmeta ) {
		      # save list meta-information to send to remote host
		      $metainfo{"$listname:$_"} = $list->{$_} if exists $list->{$_};
		  }
		  next FILE;
	      }
	  }
	  foreach my $map ( values %{$self->files(obj_type=>"map")} ) {
	      if ( $map->{path} eq $file ) {
		  $rpc->attach( "map" => [$PREFIX."/".$file, $file]  );
		  (my $mapname = $map->{path}) =~ s/.*\/([\w\-]+)$/$1/;
		  foreach  ( @mapmeta ) {
		      # save map meta-information to send to remote host
		      $metainfo{"$mapname:$_"} = $map->{$_} if exists $map->{$_};
		  }
		  next FILE;
	      }
	  }
	  foreach my $lfile ( values %{$self->files(obj_type=>"file")} ) {
	      if ( $lfile->{path} eq $file ) {
		  $rpc->attach( "file" => [$PREFIX."/".$file, $file]  )
		      if (-f $PREFIX."/".$file);
 		  next FILE;
	      }
	  }
     }
        $rpc->call("Subscription.Update", $self->{name}, $hostname, %metainfo);
        $rpc->wait();
        if ( my $err = $rpc->error() ) {
            $self->raise_error($err);
            $hosts->{$host}{status} = "$err";
        }
        else {
            my @data = $rpc->data($host);
            my ($ok, $fail);
            my $newest = 0;
            foreach ( @data ) {
                $newest = $_->{MODIFIED} if ( $_->{MODIFIED} and $_->{MODIFIED} > $newest );
		if ( exists($_->{STATUS}) and defined($_->{STATUS}) 
		     and ($_->{STATUS} eq "Ok") ) {
                    $ok++;
                }
                elsif ( $_->{STATUS} ) {
                    $fail++;
                }
            }
            $hosts->{$host}{last_updated} = $newest;
            $hosts->{$host}{status} = "";
            $hosts->{$host}{status} .= "$ok file/list/map(s) updated." if $ok;
            $hosts->{$host}{status} .= "Failed to sync $fail file/list/map(s)." if $fail;
        }
    }
    return $hosts;
}

sub sync_host {

}

sub subscribe {
    my $self = shift;
    my $host = shift;
    my %opts = @_;
    my $h = get_hosts();
    unless ( $h->{$host} ) {
        $self->raise_error("Can't find host $host in Server Groups configuration.");
        return;
    }
    if ( is_self($h->{$host}) ) {
        $self->raise_error("Publisher and requested subscriber is the same host.");
        return;
    }
    
    my $path = "publication#$self->{name}";
    _read_config(writable=>1) unless $writable;
    my $node = path_node($p->root, "$path.hosts", create => 'group');
    my @hosts = path_value($p->root, "$path.hosts");
    
    my $license = PureMessage::_sys::lic_file();
    if ( PureMessage::_sys::lic_check($license, 'PureMessage::ServerGroup::AddOn') eq "valid" ) {
        my %lic = PureMessage::_sys::lic_features($license);
        my %lic_data = PureMessage::_sys::lic_feature_kv($lic{'PureMessage::ServerGroup::AddOn'}); 
	# splits license key for the feature
        $lic_data{NumberOfHosts} ||= 1; # make test scripts happy
        if ( scalar @hosts >= $lic_data{NumberOfHosts} ) {
            $self->raise_error("Your license allows only $lic_data{NumberOfHosts} subscribers per publication.");
            return;
        }
    }
    else {
        $self->raise_error("Can't find valid license for PureMessage::ServerGroup::AddOn package.");
        return;
    }
    
    foreach ( @hosts ) {
        if ( $_ eq $host ) {
            $self->raise_error("$host is already subscribed");
            return;
        }
    }
    push @hosts, $host;
    path_value($p->root, "$path.hosts", \@hosts);
    save($self);
    $self->_init();
}

sub unsubscribe {
    my $self = shift;
    my $host = shift;
    my $path = "publication#$self->{name}";
    _read_config(writable=>1) unless $writable;
    my $node = path_node($p->root, "$path.hosts", create => 'group');
    my @hosts = path_value($p->root, "$path.hosts");
    my @new_hosts;
    foreach (@hosts) {
        push @new_hosts, $_ if $_ ne $host;
    }
    path_value($p->root, "$path.hosts", \@new_hosts);
    save($self);
    $self->_init();
    # XXX - send UNSUB RPC request
}

sub raise_error {
    my $self = shift;
    if ( $self ) {
	$self->{error} = "@_";
    }
    else {
	$error = "@_";
    }
}

sub error {
    my $self = shift;
    if ( $self ) {
	return $self->{error};
    }
    else {
	return $error;
    }
}

sub save {
    my $self = shift;
    if ( $self ) {
        path_node($p->root, "publication#$self->{name}.modified", create => 1);
        path_value($p->root, "publication#$self->{name}.modified", _today());
    }
    $p->save();
    _read_config();
}

1;
